/* $Id: glotvrecord.c 322 2008-09-17 11:55:43Z bd $
 *
 * Copyright 2004,2008 Bjorn Danielsson
 * http://glotv.dax.nu/
 *
 * This file is part of GLOTV. GLOTV is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License as
 * published at this URL: http://www.gnu.org/licenses/gpl.html.
 *
 * Usage: glotvrecord [-p pidfile] path maxsize endtime starttime
 * With no arguments it will stop a previously started glotvrecord
 */

/*
 * Turn on _GNU_SOURCE to make sure O_LARGEFILE is defined.
 */

#define _GNU_SOURCE	1

#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>

#ifndef O_LARGEFILE
#define O_LARGEFILE 0	/* for compiling on non-Linux systems */
#endif

/*
 * Edit and uncomment the following line if you want to restrict
 * the paths to which this program can write data. This is useful
 * if for some reason you need to run the program setuid in order
 * to access some exotic foreign file system. Any file opened for
 * writing must start with the string defined by ALLOWEDROOT.
 */

/* #define ALLOWEDROOT "/var/spool/glotvrecord/" */

#ifndef ALLOWEDROOT
#define ALLOWEDROOT ""
#endif

/*
 * TIMECHECKLIMIT is the number of bytes to read between time
 * checks. Since time_t has a 1-second resolution, a reasonable
 * value is something that corresponds to a few seconds at the
 * lowest bitrates.
 */

#define TIMECHECKLIMIT 300000

#define BUFSIZE 32768

char *pidfile = "/tmp/glotvrecord.pid";

typedef long long int bytecounter;

char buf[BUFSIZE];
char *progname = "";

#ifdef __FreeBSD__
/* C99 function that is missing in several FreeBSD releases: */
long long atoll(const char *str) {
  return strtoll(str, (char **)NULL, 10);
}
#endif

void fatal(char *msg) {
  if (msg)
    fprintf(stderr, "%s: %s\n", progname, msg);
  else
    perror(progname);
  exit(1);
}

int createpidfile() {
  int fd;
  if ((fd = open(pidfile, (O_CREAT | O_EXCL | O_WRONLY), 00644)) > -1) {
    sprintf(buf, "%d\n", getpid());
    write(fd, buf, strlen(buf));
    close(fd);
    return 1;
  }
  return 0;
}

int killpidfile() {
  int fd;
  if ((fd = open(pidfile, O_RDONLY)) > -1) {
    pid_t pid = 0;
    read(fd, buf, 16);
    pid = atoi(buf);
    close(fd);
    unlink(pidfile);
    if (pid && kill(pid, SIGTERM) == 0) return 1;
  }
  return 0;
}

int main(int argc, char *argv[]) {
  char *path;
  bytecounter maxsize = 0;
  time_t endtime = 0;
  time_t starttime = 0;
  int fd;
  ssize_t n;
  int writing_enabled = 0;
  bytecounter bytes_written = 0;
  size_t bytes_since_check = TIMECHECKLIMIT + 1;
  progname = argv[0];
  while (1) {
    switch (getopt(argc, argv, "p:")) {
    case -1:
      goto quit;
    case 'p':
      pidfile = optarg;
      break;
    default:
      fprintf(stderr, "%s: read source code for usage instructions\n", argv[0]);
      exit(1);
    }
  }
 quit:
  ;
  int offset = optind-1;
  argv += offset;
  argc -= offset;
  if (argc > 5) {
    fprintf(stderr, "%s: extra arguments ignored, continuing\n", argv[-offset]);
    argc = 5;
  }
  switch (argc) {
  case 5: starttime = atoll(argv[4]);
  case 4: endtime = atoll(argv[3]);
  case 3: maxsize = atoll(argv[2]);
  case 2: path = argv[1];
    break;
  default:
    if (killpidfile()) exit(0);
    exit(1);
  }
  if (strncmp(path, ALLOWEDROOT, sizeof(ALLOWEDROOT)-1) != 0)
    fatal("path is not within allowed root");

  if ((fd = open(path, (O_CREAT | O_WRONLY | O_TRUNC | O_LARGEFILE), 00666)) < 0)
    fatal(NULL);

  if (!createpidfile()) {
    if (!killpidfile()) exit(1);
    if (!createpidfile()) exit(1);
  }

  // return 0 when a segment is filled, 1 for all other cases.
  int exit_code = 0;

  while (1) {
    if (bytes_since_check > TIMECHECKLIMIT) {
      time_t now = time(NULL);
      if (endtime && now >= endtime) {
	exit_code = 1;
	break;
      }
      if (now >= starttime) {
	writing_enabled = 1;
      }
      bytes_since_check = 0;
    }
    n = read(0, buf, BUFSIZE);
    if (n < 0) fatal(NULL);
    else if (n == 0) {
      fprintf(stderr, "%s: premature end of recording due to EOF\n", argv[-offset]);
      exit_code = 1;
      break;
    } else if (writing_enabled) {
      char *p = buf;
      size_t k = n;
      while (k > 0) {
	ssize_t m = write(fd, p, k);
	if (m < 0) fatal(NULL);
	p += m;
	k -= m;
      }
      bytes_written += n;
      if (maxsize > 0 && bytes_written + BUFSIZE > maxsize) break;
    }
    bytes_since_check += n;
  }
  close(fd);
  unlink(pidfile);
  exit(exit_code);
}

